# SOURCE CODE partly from Christoph Haenle
#
# (c) 1999-2000 Aberdeen University Electronics Research Lab
#
# THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
#
#    Modified by: Martin Koyabe, koyabe@erg.abdn.ac.uk
#           File: mftp_sim_general.tcl
#    Last update: February 21 2001
#
# This script creates (1) source and (N) receiver nodes with asymmetric links.
# Each downlink experiences both random packet errors and losses (generated by error model). 



# Get input values

if {$argc != 3} {
    puts stderr {usage: ns mftp_sim_general.tcl <GroupSize: 1,8,16,32..> <Number_of_receivers> <File Size>}
    exit 1
}

# finish simulation routine

proc finish { } {
    global ns sender mftp_trace number_of_receivers
    $ns flush-trace
    #close $nf
    puts stdout "view output in mftp_sim_general.tr"
    puts stdout "Simulation successfully completed ...."
    exit 0
}

# setup simulator
set ns [new Simulator -multicast on]

# Set virtual classifier on
# Simulator set node_factory_ VirtualClassifierNode

set verbose 1
set group_size [lindex $argv 0]
set number_of_receivers [lindex $argv 1]
set file_size [lindex $argv 2]
set group_addr [Node allocaddr]	;# allocate multicast group address


# Trace link events

set mftp_trace [open /tmp/mftp_sim_general.tr w]

# create the nam output file
# set nf [open mftp_sim_general.nam w]
# $ns namtrace-all $nf

# Create Error Model
#  states:      an array of states (error models). 
#  periods:     an array of average state durations. 
#  probability: an array of state probability.
#  transunit:   one of [pkt|byte|time]. 
#  sttype:      type of state transitions to use: either time or pkt. 

#for {set k 0} {$k < $number_of_receivers} {incr k} {
#   set error_model($k) [new ErrorModel]
#    $error_model($k) unit bit
#    $error_model($k) set rate_ 0.000001
#    $error_model($k) ranvar [new RandomVariable/Uniform]
#    $error_model($k) drop-target [new Agent/Null]
#}

#set tmp  [new ErrorModel/Uniform 0.0000 bit]
#set tmp1 [new ErrorModel/Uniform 1.0000 bit]

# array of states (error models)
#set m_states [list $tmp $tmp1]

# average fade periods for each of the states, tmp, tmp1 and tmp2, respectively
#set m_periods [list [expr 11*3600] 1500]

# probability distribution of fades
#set m_probability [list 0.95 0.05]

# specify transmission unit either 'pkt', 'bit', 'time'
#set m_trunit pkt

# Use time-based transition
#set m_sttype time

# assign error model to a variable
#for {set k 0} {$k < $number_of_receivers} {incr k} {
#        set error_model($k) [new ErrorModel/RandomMultiState $m_states $m_periods $m_probability $m_trunit $m_sttype]
#}

set rng [new RNG]
$rng seed 0

set range [list 0 100]
set duration_range [list 70.0 20.0 10.0 6.0 4.0 2.75 2.0 1.75 1.5 1.0]
set s_time 0

for {set k 0} {$k < $number_of_receivers} {incr k} {

     set time [expr [expr [$rng next-random]] % [expr [lindex $range 1]/10]]
     while {$time < [lindex $range 0]} {
           set time [expr [expr [$rng next-random]] % [lindex $range 1]]
    }

#    puts "start time is: $time"

    set f_start($k) [list $time]
    set f_duration($k) [list [lindex $duration_range [expr [$rng integer 10]]]]

    set error_model_($k) [new ErrorModel/TimeSeries $f_duration($k) $f_start($k) $range $s_time]
    $error_model_($k) unit pkt
    $error_model_($k) set rate_ 0.9
    $error_model_($k) drop-target [new Agent/Null]
}


# create network topology like this:
#                          2
#                         /.
#                        / .
#              0--------1  .
#                        \ .
#                         \.
#                          n


# setup nodes

for {set i 0} {$i < [expr $number_of_receivers+2]} {incr i} {
      set node_($i) [$ns node]
}


# Creating links for WAN

if {$verbose} { puts "Creating links for WAN ...."; flush stdout }
$ns duplex-link $node_(0) $node_(1)  2500Kb 0ms DropTail


# $ns simplex-link $node_(1) $node_(0)  2500Kb 0ms DropTail

$ns queue-limit $node_(0) $node_(1)  10000000
$ns queue-limit $node_(1) $node_(0)  10000000

# getting the queue details

# set queue01 [[$ns link $node_(0) $node_(1)] queue]
# set queue10 [[$ns link $node_(1) $node_(0)] queue]

# puts "queue size node(0)-node(1): [$queue01 set limit_]"
# puts "queue size node(1)-node(0): [$queue10 set limit_]"

# Creating links for MAN with random delays


if {$verbose} { puts -nonewline "Creating links for LAN with (1..$number_of_receivers) nodes "; flush stdout }
for {set m 2} {$m < [expr $number_of_receivers+2]} {incr m} {
    $ns simplex-link $node_(1) $node_($m)  2500Kb 250ms DropTail
    $ns simplex-link $node_($m) $node_(1)    64Kb 250ms DropTail
    puts -nonewline "."; flush stdout
}
puts " "; flush stdout

# DM set PruneTimeout 0.5
# set mrthandle [$ns mrtproto DM {}]
# set mrthandle [$ns mrtproto CtrMcast {}]

DM set PruneTimeout 0.3         ;  # default 0.5 (sec)
DM set CacheMissMode dvmrp      ;  # default pimdm
$ns mrtproto DM

# CtrMcast


# Insert link errors

 if {$verbose} { puts -nonewline "Inserting link errors (1..$number_of_receivers) nodes "; flush stdout }
 for {set l 2} {$l < [expr $number_of_receivers+2]} {incr l} {
     $ns link-lossmodel $error_model_([expr $l-2]) $node_(1) $node_($l)
     puts -nonewline "."; flush stdout
 }
 puts " "; flush stdout


# color codes
$ns color 0 black
$ns color 1 blue
$ns color 2 purple
$ns color 3 green


# setup MFTP sender
set sender [new Agent/MFTP/Snd];        # new sender instance 
$sender set dst_addr_ $group_addr;      # set destination address
$sender set fid_ 2;                     # flow-id (for use in nam)
$sender set fileSize_ $file_size;       # size of file in bytes
$sender set dtusPerGroup_ $group_size;  # set transmission group size (in pkts/group)
$sender set dtusPerBlock_ 1424;         # set block size (in pkts/block), must be multiple of 8
$sender set dtuSize_ 512;               # set packet size for transmission
$sender set interval_ 0.0016384         # 0.0016384 for 512, 0.0047104 for 1472- set interval between 2 packet transmissions (in s)
$sender trace $mftp_trace;              # optional set trace-file
$ns attach-agent $node_(0) $sender;     # attach MFTP-sender to node

set sender_addr  [$sender set agent_addr_]
set sender_port  [$sender set agent_port_]

# setting up MFTP receivers

for {set j 2} {$j < [expr $number_of_receivers+2]} {incr j} {
    set recv($j) [new Agent/MFTP/Rcv/Stat]
    $recv($j) set dst_addr_ $group_addr
    $recv($j) set reply_addr_ $sender_addr;  # address to which packets are sent (unicast)
    $recv($j) set reply_port_ $sender_port;  # port to which packets are sent (unicast) 
    $recv($j) set fid_ 2
    $recv($j) set fileSize_ $file_size
    $recv($j) set dtusPerGroup_ $group_size
    $recv($j) set dtusPerBlock_ 1424
    $recv($j) set dtuSize_ 512;              # set packet size for transmission
    $recv($j) trace $mftp_trace
    $ns attach-agent $node_($j) $recv($j)
    $ns at 0.1 "$recv($j) start"
}

puts "Transmitting $file_size byte file with group-size $group_size"; flush stdout
puts "simulation is running..."; flush stdout
$ns at 0.0 "$sender start";             # set time of start of transmission
# $ns at 99999.99 "[$ns clearMemTrace]"
# $ns at 1000000.0 "exit 0"
$ns at 1000000.0 "finish"

$ns run





