# SOURCE CODE partly from Christoph Haenle
#
# (c) 1999-2000 Aberdeen University Electronics Research Lab
#
# THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
#
#    Modified by: Martin Koyabe, koyabe@erg.abdn.ac.uk
#           File: mftp_sim_2.tcl
#    Last update: Jan 24, 2000
#
# This script creates a source and two receivers nodes with asymmetric links.
# Each downlink experiences random packet errors/losses (generated by error model) and 
# losses. The script computes the throughput achieved at each receiver.


# finish simulation routine

proc finish { } {
    global ns sender mftp_trace 
    $ns flush-trace

    puts stdout "view output in mftp_sim_3.tr"
    puts stdout " "
    puts stdout "Simulation successfully completed ...."
    exit 0
}



# setup simulator

set ns [new Simulator -multicast on]

set verbose 1
set group_size 16
set file_size 2000000
set group_addr [Node allocaddr]	;# allocate multicast group address


# Trace link events

set mftp_trace [open mftp_sim_3.tr w]

# create the nam output file
# set nf [open mftp_2.nam w]
# $ns namtrace-all $nf


# Create Error Model

for {set k 1} {$k <= 3} {incr k} {
    set error_model($k) [new ErrorModel]
    $error_model($k) unit pkt
    $error_model($k) set rate_ 0.01
    $error_model($k) ranvar [new RandomVariable/Uniform]
    $error_model($k) drop-target [new Agent/Null]
}


# create network topology like this:
#                          3
#                         /
#                        /
#              0--------1
#                        \
#                         \
#                          2


# setup nodes

set i 0
while {$i < 5} {
      set node_($i) [$ns node]
      incr i
}


# Creating links for WAN 0

if {$verbose} { puts "Creating links for WAN ...."; flush stdout }

$ns simplex-link $node_(0) $node_(1)  2500Kb 0ms DropTail
$ns simplex-link $node_(1) $node_(0)    64Kb 0ms DropTail

# Creating links for MAN 0

if {$verbose} { puts "Creating links for MAN/LAN ...."; flush stdout }

$ns simplex-link $node_(1) $node_(2)  2500Kb 250ms DropTail
$ns simplex-link $node_(2) $node_(1)    64Kb 250ms DropTail
$ns simplex-link $node_(1) $node_(3)  2500Kb 250ms DropTail
$ns simplex-link $node_(3) $node_(1)    64Kb 250ms DropTail
$ns simplex-link $node_(1) $node_(4)  2500Kb 250ms DropTail
$ns simplex-link $node_(4) $node_(1)    64Kb 250ms DropTail


# Insert link errors

if {$verbose} { puts "Inserting link errors ...."; flush stdout }

$ns link-lossmodel $error_model(3) $node_(1) $node_(4)
#$ns link-lossmodel $error_model(2) $node_(1) $node_(3)
#$ns link-lossmodel $error_model(1) $node_(1) $node_(2)


# color codes
$ns color 0 black
$ns color 1 blue
$ns color 2 purple
$ns color 3 green

#DM set PruneTimeout 0.5 	;# PruneTimeout set to 0.5s
#set mrthandle [$ns mrtproto DM {}]

set mproto CtrMcast
set mrthandle [$ns mrtproto $mproto {}]


# setup MFTP sender

set sender [new Agent/MFTP/Snd];        # new sender instance
$ns attach-agent $node_(0) $sender;     # attach MFTP-sender to node
$sender set dst_addr_ $group_addr;      # set destination address
$sender set fid_ 1;                     # flow-id (for use in nam)
$sender set fileSize_ $file_size;       # size of file in bytes
$sender set dtusPerGroup_ $group_size;  # set transmission group size (in pkts/group)
$sender set dtusPerBlock_ 1472;         # set block size (in pkts/block), must be multiple of 8
$sender set dtuSize_ 512;               # set packet size for transmission
$sender set interval_ 0.002;            # set interval between 2 packet transmissions (in s)
$sender trace $mftp_trace;              # optional set trace-file


set sender_addr [$sender set dst_addr_]
set total_nodes 5

# setting up MFTP receivers

for {set j 2} {$j < $total_nodes} {incr j} {
    set recv($j-1) [new Agent/MFTP/Rcv]
    $ns attach-agent $node_([expr ($total_nodes+1)-($j)]) $recv($j-1)
    $recv($j-1) set dst_addr_ $group_addr
    $recv($j-1) set reply_addr_ $sender_addr;  # address to which packets are sent (unicast) 
    $recv($j-1) set fid_ 2
    $recv($j-1) set fileSize_ $file_size
    $recv($j-1) set dtusPerGroup_ $group_size
    $recv($j-1) set dtusPerBlock_ 1472
    $recv($j-1) trace $mftp_trace
    $ns at 0.1 "$recv($j-1) start"
}

puts stdout "Transmitting $file_size byte file with group-size $group_size"
puts stdout "simulation is running..."
$ns at   0.0 "$sender start";            # set time of start of transmission

$ns at 1000.0 "finish"
$ns run