#puts stdout $argv


proc ldelete { list value } {
    set ix [lsearch -exact $list $value]
    if { $ix >= 0 } {
        return [lreplace $list $ix $ix]
    } else {
        return $list
    }
}

proc Incr { varName {amount 1}} {
    upvar $varName var
    if [info exists var] {
        set var [expr $var+$amount]
    } else {
        set var $amount
    }
    return $var
}

if {[llength $argv] == 0} {
    puts stderr "Usage: tclsh EvaluateMFTPTrace.tcl tracefile"
    exit 0
}

if [catch { open [lindex $argv 0] r} inp_fd] {
    puts stderr "Can not open file [lindex $argv 0] for read"
    exit 0
}

set act_nb_recvs 0
set act_recv_addrs {}
set recv_addrs {}


puts stderr "Scanning for MFTP receivers..."

while { [gets $inp_fd line] >= 0 } {
    if { [scan $line "%s %s recv-useful %lu %lu %s" \
              time addr CurrentPass CurrentGroup Codeword] == 5 && \
         ![info exists receive($addr,last-packet)] } {
        lappend recv_addrs $addr
        set receive($addr,last-packet) NONE
        set receive($addr,longest-burst) -1
    }
}

set nb_recvs [llength $recv_addrs]

puts stderr "$nb_recvs receiver(s) found with addresses: $recv_addrs"
seek $inp_fd 0 start   ;# rewind


set act_recv_addrs $recv_addrs               ;# list of active receivers
set act_nb_recvs [llength $act_recv_addrs]   ;# number of active receivers,
                                              # covers initially all

set send(packetlist) "" ;# the list "send(packetlist)" caches all packets (while
# processing the tracefile line-by-line) which were sent, but not yet received
# (or detected as lost) by all receivers, i.e. it sort of holds the packets
# currently travelling through the network.

# loop over lines in file
while { [gets $inp_fd line] >= 0 } {
    if { [scan $line "%s %s %s %lu %lu %s" \
              time addr command CurrentPass CurrentGroup Codeword] != 6 } {
    } else {
        set packet "$CurrentPass:$CurrentGroup"
        if { $command == "send-notify" } {
            lappend send(packetlist) $packet
            set send($packet,ref-cnt) $act_nb_recvs
            set send($packet,time) $time
        } elseif { [lsearch -exact $act_recv_addrs $addr] >= 0 && \
                       [regexp "recv-(useful|dependent|(group-full))" $command] } {
            set start [lsearch -exact $send(packetlist) $receive($addr,last-packet)]
            set fin   [lsearch -exact $send(packetlist) $packet]
            # note: if start is -1, then incr sets it to 0, and this is just what we want
            incr start
            
            if { $fin == -1 } {
                error("Internal bug in script file, please contact author chris@cs.vu.nl");
            }
            foreach pkt [lrange $send(packetlist) $start $fin] {
                incr send($pkt,ref-cnt) -1
                if { $send($pkt,ref-cnt) == 0 } {
                    set send(packetlist) [ldelete $send(packetlist) $pkt]
                    unset send($pkt,ref-cnt)
                    unset send($pkt,time)
                }
            }
            set receive($addr,last-packet) $packet
            set burstlen [expr $fin - $start]
            set index "$addr,burstloss,$burstlen"
            if { $burstlen <= $receive($addr,longest-burst) } {
                incr receive($index)
            } else {
                for { set i [expr $receive($addr,longest-burst)+1] } { $i < $burstlen } { incr i } {
                    set receive($addr,burstloss,$i) 0
                }
                set receive($index) 1
                set receive($addr,longest-burst) $burstlen
            }
            switch -- $command {
                {recv-useful} { Incr receive($addr,nb_useful_recv) }
                {recv-group-full} { Incr receive($addr,nb_full_disc) }
                {recv-dependent} { Incr receive($addr,nb_lin_dep_disc) }
            }
        } elseif { [lsearch -exact $act_recv_addrs $addr] >= 0 && \
                       $command == "done-notify" } {
            scan $line "%s %s %s %lu %lu %s" \
                time addr command CurrentPass CurrentGroup Codeword
            set start [lsearch -exact $send(packetlist) $receive($addr,last-packet)]
            set fin   [expr [llength $send(packetlist)]-1]
            # note: if start is -1, then incr sets it to 0, and this is just what we want
            incr start
            if { $start <= $fin } {
                foreach pkt [lrange $send(packetlist) $start $fin] {
                    incr send($pkt,ref-cnt) -1
                    if { $send($pkt,ref-cnt) == 0 } {
                        set send(packetlist) [ldelete $send(packetlist) $pkt]
                        unset send($pkt,ref-cnt)
                        unset send($pkt,time)
                    }
                }
            }
            set act_recv_addrs [ldelete $act_recv_addrs $addr]
            incr act_nb_recvs -1
        }
    }
}

puts stdout "Number of consecutive losses"
puts stdout "Each row belongs to one receiver and the i-th number (zero-based) shows the"
puts stdout "number of burst losses of length i (where 0 = no loss)"
foreach addr $recv_addrs {
    puts -nonewline stdout "Receiver $addr: "
    for { set i 0 } { $i <= $receive($addr,longest-burst) } { incr i } {
        puts -nonewline stdout "$receive($addr,burstloss,$i),"
    }
    
    if { [info exists receive($addr,nb_useful_recv)] } {
        puts stdout "nb-useful-recv:$receive($addr,nb_useful_recv),\
                     nb-full-disc:$receive($addr,nb_full_disc),\
                     nb_lin-dep-disc:$receive($addr,nb_lin_dep_disc)"     ;# newline
    } else {
        puts stdout "Warning: receiver $addr did not finish reception!"
    }
}
