
#-------------------------------------------------------------------------
#  Originally provided by Chris Hanle
#
#  Modification has been done to suit my simulation
#
#  tracelook-mftp scans the trace file generated by MFTP-ns and computes
#  number of receivers, receiver addresses, burst loss, useful packets 
#  received, packets discarded due to linear dependency, packets discarded
#  due to group-full, good throughput and throughput efficiency between the
#  source and the receivers
#
#  Contributed by: Martin Koyabe, koyabe@erg.abdn.ac.uk
#       File name: tracelook-mftp.tcl 
#     Last update: 25-Jan-2000
# ------------------------------------------------------------------------



# Procedure delete list

proc ldelete {list value} {
     set ix [lsearch -exact $list $value]
     
     if {$ix >= 0} {
        return [lreplace $list $ix $ix]
     } else {
        return $list
     }
}


# Procedure for increamenting amount

proc Incr {varName {amount 1}} {
     upvar $varName var
     if [info exists var] {
        set var [expr $var+$amount]
     } else {
        set var $amount
     }     
     return $var
}


# program starts here

if {$argc != 4} {
    puts stderr {Usage: tclsh tracelook-mftp.tcl <trace-file name> <file size [Bytes]> <dtu-size [Bytes]> <bandwidth [Kb/s]>}
   exit 0
}


if [catch {open [lindex $argv 0] r} inp_fd] {
   puts stderr "ERROR: Opening file [lindex $argv 0]"
   exit 0 
}

# Create the output files  

if [catch {open /tmp/tracelook_complete.tr a+ } outp_complete] {
   puts stderr "Cannot open $outp_file: $outp_complete"
   exit 0
}

if [catch {open /tmp/tracelook_goodput.tr a+} outp_goodput] {
   put stderr "Cannot open $outp_file: $outp_goodput"
   exit 0
}

if [catch {open /tmp/tracelook_useful.tr a+ } outp_useful] {
   puts stderr "Cannot open $outp_file: $outp_useful"
   exit 0
}

if [catch {open /tmp/tracelook_duplicate.tr a+ } outp_duplicate] {
   puts stderr "Cannot open $outp_file: $outp_duplicate"
   exit 0
}

if [catch {open /tmp/tracelook_cum_lossrate.tr a+ } outp_c_lossrate] {
   puts stderr "Cannot open $outp_file: $outp_c_lossrate"
   exit 0
}


if [catch {open /tmp/tracelook_bandwidth.tr a+ } outp_bandwidth] {
   puts stderr "Cannot open $outp_file: $outp_bandwidth"
   exit 0
}

if [catch {open /tmp/data_efficiency.tr a+ } outp_fd] {
   puts stderr "Cannot open $outp_file: $outp_fd"
   exit 0
}

if [catch {open /tmp/data_packet.tr a+ } outp_fd2] {
   puts stderr "Cannot open $outp_file: $outp_fd2"
   exit 0
}

if [catch {open /tmp/completion_summary.tr a+ } outp_completion] {
   puts stderr "Cannot open $outp_file: $outp_completion"
   exit 0
}


set fileSize [lindex $argv 1]	;# set file size to be transferred
set dtuSize [lindex $argv 2]	;# should be the same as dtuSize in tcl scripts
set link_speed [lindex $argv 3] ;# set bandwidth of link

set act_nb_recvs 0	;# initialize active number of receivers
set act_recv_addrs { }  ;# initialize active receiver address array list
set recv_addrs { }      ;# initialize receiver address array list
set receiver_addrs { }
set interval_counts { } ;# initialize interval counts array
set int_counts { }

set real_complete [expr [expr double($fileSize)*8.0]/[expr $link_speed*1000.0]]	;# compute the estimated time of completion based on aapproximation
set time_step 0.05
set transfer_time_interval $time_step
set old_transfer_time_interval 0

# While loop for scanning MFTP tracefile for receivers

puts stderr "Scanning for MFTP receivers....."


set complete_recvs_count 0
set tot_complete_recvs_count 0
set cum_percent_complete 0
set count 0
set begin_flag 0
set transmit_start_time 0.0

set t_efficiency 0
set t_useful_lost 0
set t_useful_recv 0
set t_full_disc 0
set t_lin_dep_disc 0
set t_useful_sent 0
set t_packets_sent 0
set t_send_status_req 0
set t_bandwidth 0
set t_done_notify 0
set t_total_packets 0

set tt_done_notify 0
set tt_useful_lost 0
set tt_useful_recv 0
set tt_full_disc 0
set tt_lin_dep_disc 0
set tt_useful_sent 0
set tt_packets_sent 0
set tt_send_status_req 0
set tt_bandwidth 0

set start_flag 0
set t_packet "0:0"

while { [gets $inp_fd line] >= 0 } {
    if { [scan $line "%s %s recv-useful %lu %lu %s" \
              time addr CurrentPass CurrentGroup Codeword] == 5 && \
         ![info exists receiver($addr,last-packet)] } {
        lappend receiver_addrs $addr
        set receiver($addr,last-packet) NONE
        set receiver($addr,longest-burst) -1
    }
}

set nb_recvs [llength $receiver_addrs]
set act_nb_recvs $nb_recvs
set cnt_packet_lost $nb_recvs
set cnt_useful 0
set useful_lost 0

# Output numbers of receivers found
puts stderr "$nb_recvs receiver(s) found with addresses: $receiver_addrs"
puts stdout "Writing output file..."

seek $inp_fd 0 start   ;# rewind
set send(packetlist) " " ; # the list send(packetlist) caches all ppackets sent

while {[gets $inp_fd line] >= 0} {
      if {[scan $line "%s %s %s %lu %lu %s" \
         time addr command CurrentPass CurrentGroup Codeword] == 6} {

         set packet "$CurrentPass:$CurrentGroup"

         if {$command == "recv-useful"} { 
  
             if {![info exists receive($addr,last-packet)]} {

                if {![info exists transmit_data($addr,flag)]} {
          
                   set transmit_data($addr,start_time) $time
                   set transmit_data($addr,flag) 1
                }

                lappend recv_addrs $addr
                set receive($addr,last-packet) NONE
                set receive($addr,longest-burst) -1
	     }

            incr t_useful_recv 1

            if { ![regexp $packet $t_packet] } {
               set useful_lost [expr $act_nb_recvs - $cnt_useful]
               set t_useful_lost [expr $t_useful_lost+$useful_lost]
               set cnt_useful 0
               set t_packet $packet
#               puts stdout "lost packets are: $t_useful_lost"
            }

            incr cnt_useful 1

         } elseif {$command == "recv-group-full"} {
             incr t_full_disc
         } elseif {$command == "recv-dependent"} {
            incr t_lin_dep_disc
         } elseif {$command == "send-status-req"} {
            incr t_send_status_req
         } 

         if {$command == "done-notify" && $transmit_data($addr,flag) == 1} {
             set transmit_data($addr,stop_time) $time
             set transmit_data($addr,transfer_time) $transmit_data($addr,stop_time)
             incr complete_recvs_count               
             set transmit_data($addr,flag) 0      
             incr t_done_notify
             set act_nb_recvs [expr $nb_recvs - $t_done_notify]
         } 

         if {$command == "send-notify"} {
            incr t_useful_sent
         } 

         if {$time >= $transfer_time_interval} {

	    if { $count == 0 } {
               set time_interval $time 
               set old_transfer_time_interval $transfer_time_interval
            } else {
               set time_interval  [expr $time-$old_transfer_time_interval]
               set old_transfer_time_interval $transfer_time_interval
	    }

            incr count
            lappend int_counts $count
            set interval_count($count,complete_receivers) $complete_recvs_count
            set interval_count($count,interval_time) $transfer_time_interval
            set tot_complete_recvs_count [expr $tot_complete_recvs_count+$complete_recvs_count]
            set complete_recvs_count 0      
            set transfer_time_interval [expr $transfer_time_interval+$time_step]   

            set t_packets_sent [expr $t_useful_sent*$nb_recvs]

            set tt_useful_lost  [expr $tt_useful_lost+$t_useful_lost]     
            set tt_useful_recv  [expr $tt_useful_recv+$t_useful_recv]
            set tt_useful_sent  [expr $tt_useful_sent+$t_useful_sent]
            set tt_packets_sent [expr $tt_packets_sent+$t_packets_sent]
            set tt_lin_dep_disc [expr $tt_lin_dep_disc+$t_lin_dep_disc]
            set tt_full_disc    [expr $tt_full_disc+$t_full_disc]

            if {$t_done_notify <= $nb_recvs} {

               puts $outp_useful "[format "%7.5f" $time]  $t_useful_recv" 
               puts $outp_duplicate "[format "%7.5f" $time] $t_full_disc" 
           

#               puts "sent packets: $t_packets_sent lost packets: $t_useful_lost"
           
               if {$t_packets_sent > 0} {
                  set  t_goodput [expr double($t_useful_recv)/double($t_packets_sent)]
		  if {$t_goodput != 0} {
                     puts $outp_goodput "[format "%7.5f" $time], [format "%2.4f" $t_goodput]"
                  }
               }
            
               if {$t_useful_recv > 0} {
                  set t_bandwidth [expr double($t_packets_sent)/double($t_useful_recv)]
		  if {$t_bandwidth != 0} {
                     puts $outp_bandwidth "[format "%7.5f" $time] [format "%3.4f" $t_bandwidth]"
                  }
	       } 

#            puts "sent packets: $t_useful_sent received: $t_useful_recv"
#            puts "lost packets are: $t_useful_lost Time interval: $time_interval"

               if {$t_packets_sent > 0} { 
                  set t_lossrate  [expr double($tt_useful_lost)*100.0/double($tt_packets_sent)]
                  puts $outp_c_lossrate "[format "%7.5f" $time] [format "%2.4f" $t_lossrate]"
	       } 
                  
	    }
 
            set t_useful_sent 0
	    set t_useful_lost 0
            set t_useful_recv 0
            set t_full_disc 0
            set t_lin_dep_disc 0
            set t_packets_sent 0
	    set t_bandwidth 0
            set t_total_packets 0

         }
     }
}

puts $outp_c_lossrate  " "
puts $outp_useful  " "
puts $outp_duplicate  " "
puts $outp_bandwidth " "
puts $outp_goodput " "

incr count
lappend int_counts $count
set interval_count($count,complete_receivers) $complete_recvs_count
set interval_count($count,interval_time) $transfer_time_interval
set tot_complete_recvs_count [expr $tot_complete_recvs_count+$complete_recvs_count]

# write complete data information
set percent_quarter 0

foreach count $int_counts {
    set interval_count($count,percent_complete) [expr double($interval_count($count,complete_receivers))*100/$nb_recvs]
    set cum_percent_complete [expr $cum_percent_complete+$interval_count($count,percent_complete)]

    if {$interval_count($count,interval_time) <= [expr $real_complete*100.0] && $interval_count($count,complete_receivers) != 0} {

	if {$cum_percent_complete >= $percent_quarter} {
	   puts $outp_complete "[format "%7.5f" $interval_count($count,interval_time)]  [format "%3.2f" $cum_percent_complete]"
    	   puts stdout "pass count: $count, time: [format "%7.5f" $interval_count($count,interval_time)], Receivers complete: \
             $interval_count($count,complete_receivers), % complete: [format "%3.2f" $interval_count($count,percent_complete)],\
             Cumulative % complete: [format "%3.2f" $cum_percent_complete]"
    	   puts $outp_completion "pass count: $count, time: [format "%7.5f" $interval_count($count,interval_time)], Receivers complete: \
             $interval_count($count,complete_receivers), % complete: [format "%3.2f" $interval_count($count,percent_complete)],\
             Cumulative % complete: [format "%3.2f" $cum_percent_complete]"
           set percent_quarter [expr ($percent_quarter + 10)]
        }
    }
}

puts $outp_complete " "

# Reset the pointer to the begining of tracefile

seek $inp_fd 0 start  ;# rewind
set act_recv_addrs $recv_addrs			;# list of active receivers
set act_nb_recvs [llength $act_recv_addrs]	;# number of active receivers, 
						 # covers initially all
				 
set send(packetlist) " "			

# The list "send(packetlist)" caches all packets (while processing the tracefile 
# line-by-line) which were sent, but not yet received (or detected as lost) by 
# all receivers, i.e. it sort of holds the packets currently travelling through
# the network.


# loop over lines in file

while {[gets $inp_fd line] >= 0} {
      if {[scan $line "%s %s %s %lu %lu %s" \
          time addr command CurrentPass CurrentGroup Codeword] != 6} {
      } else {
          set packet "$CurrentPass:$CurrentGroup"

        if {$command == "send-notify"} {
           lappend send(packetlist) $packet
           set send($packet,ref-cnt) $act_nb_recvs
           set send($packet,time) $time         

        } elseif {
           [lsearch -exact $act_recv_addrs $addr] >= 0 && \
           [regexp "recv-(useful|dependent|(group-full))" $command]} {
           
           set start [lsearch -exact $send(packetlist) $receive($addr,last-packet)]
           set fin   [lsearch -exact $send(packetlist) $packet]
           
           # Note: if start is -1, then incr sets it to 0, and this is just what we want
           
           incr start
           
           if {$fin == -1} {
              error("Internal bug in script file, please contact Chris@cs.vu.nl or koyabe@erg.abdn.ac.uk");
           }
           
           foreach pkt [lrange $send(packetlist) $start $fin] {
                   incr send($pkt,ref-cnt) -1
                   if {$send($pkt,ref-cnt) == 0} {
                      set send(packetlist) [ldelete $send(packetlist) $pkt]
                      unset send($pkt,ref-cnt)
                      unset send($pkt,time)
                   }
           }
           
           set receive($addr,last-packet) $packet
           set burstlen [expr $fin - $start]
           set index "$addr,burstloss,$burstlen"
           
           if {$burstlen <= $receive($addr,longest-burst)} {
              incr receive($index)
           } else {
              for {set i [expr $receive($addr,longest-burst)+1]} {$i < $burstlen} {incr i} {
                  set receive($addr,burstloss,$i) 0
              }
              
              set receive($index) 1
              set receive($addr,longest-burst) $burstlen
           }
           
           switch -- $command {
                  {recv-useful}     {Incr receive($addr,nb_useful_recv)}
                  {recv-group-full} {Incr receive($addr,nb_full_disc)}
                  {recv-dependent}  {Incr receive($addr,nb_lin_dep_disc)}	
           }
            
       } elseif {
          [lsearch -exact $act_recv_addrs $addr] >= 0 && $command == "done-notify"} {
          
          scan $line "%s %s %s %lu %lu %s" time addr command CurrentPass CurrentGroup Codeword
          set start [lsearch -exact $send(packetlist) $receive($addr,last-packet)]
          set fin   [expr [llength $send(packetlist)]-1]
          
          # Note: if start is -1, then incr sets it to 0, and this is just what we want
          
          incr start
          if {$start <= $fin} {
             foreach pkt [lrange $send(packetlist) $start $fin] {
                     incr send($pkt,ref-cnt) -1
                     if {$send($pkt,ref-cnt) == 0} {
                        set send(packetlist) [ldelete $send(packetlist) $pkt]
                        unset send($pkt,ref-cnt)
                        unset send($pkt,time)
                     }
             }
          }
          
          set act_recv_addrs [ldelete $act_recv_addrs $addr]
          incr act_nb_recvs -1   
      }
     }
   }
    
    puts stdout "----------------------------"
#    puts stdout "Number of Consecutive losses"
#    puts stdout "Each row belongs to one receiver and the i-th number (zero-based) shows the"
#    puts stdout "number of burst losses of length i (where 0 = no loss)"
    
    set tot_efficiency   0
#    set tot_throughput   0
    set tot_useful_recv  0
    set tot_full_disc    0
    set tot_lin_dep_disc 0

    set max_temp_efficiency  0
    set min_temp_efficiency  0
#    set max_temp_throughput  0
#    set min_temp_throughput  0

    set max_efficiency   0
    set min_efficiency   0
#    set max_throughput   0
#    set min_throughput   0
    set efficiency_flag  0
#    set throughput_flag  0
    
    if {[expr $fileSize%$dtuSize] > 0} {
       set exact_nb_useful [expr [expr $fileSize/$dtuSize]+1]
    }
    
    if {[expr $fileSize%$dtuSize] == 0} {
       set exact_nb_useful [expr $fileSize/$dtuSize]
    }
    


    
    foreach addr $recv_addrs {
            set useful_recv 0
            set lin_dep_disc 0
            set full_disc 0
            
#            puts -nonewline stdout "Receiver $addr:"
            
            for {set i 0} {$i <= $receive($addr,longest-burst)} {incr i} {
#-                puts -nonewline stdout "$receive($addr,burstloss,$i),"                     
            }
            
            if {[info exists receive($addr,nb_useful_recv)]} {
#-               puts -nonewline stdout "nb-useful-recv:$receive($addr,nb_useful_recv)"
               set useful_recv $receive($addr,nb_useful_recv)
               set tot_useful_recv [expr $tot_useful_recv+$receive($addr,nb_useful_recv)]
               
               if {[info exists receive($addr,nb_lin_dep_disc)]} {
#-                  puts -nonewline stdout " nb-lin-dep-disc:$receive($addr,nb_lin_dep_disc)"
                  set lin_dep_disc $receive($addr,nb_lin_dep_disc)
                  set tot_lin_dep_disc [expr $tot_lin_dep_disc+$receive($addr,nb_lin_dep_disc)]
               }
               
               if {[info exists receive($addr,nb_full_disc)]} {
#-                  puts -nonewline stdout " nb-full-disc:$receive($addr,nb_full_disc)"
                  set full_disc $receive($addr,nb_full_disc)
                  set tot_full_disc [expr $tot_full_disc+$receive($addr,nb_full_disc)]
               }   
               
               set tot_bytes_recv [expr $tot_useful_recv + $tot_lin_dep_disc + $tot_full_disc]              
#               set transmit_data($addr,transfer_time) [expr $transmit_data($addr,stop_time) - $transmit_data($addr,start_time)]
               set transmit_data($addr,efficiency) [expr double($receive($addr,nb_useful_recv)) / [expr $lin_dep_disc+$useful_recv+$full_disc]]
               set tot_efficiency [expr $tot_efficiency + $transmit_data($addr,efficiency)]
               set avg_efficiency [expr double($tot_efficiency) / $nb_recvs]               

#               set transmit_data($addr,throughput) [expr ($dtuSize*$receive($addr,nb_useful_recv)*8.0) / (1000*$transmit_data($addr,transfer_time))]             
#               set tot_throughput [expr $tot_throughput + $transmit_data($addr,throughput)]
#               set avg_throughput [expr double($tot_throughput) / $nb_recvs]
      
#               puts -nonewline stdout " Throughput(kb/s): [format "%5.2f" $transmit_data($addr,throughput)]"
#-               puts -nonewline stdout " Efficiency: [format "%3.4f" $transmit_data($addr,efficiency)]"
               
               if {$transmit_data($addr,efficiency) > $max_temp_efficiency} {
                  set max_efficiency $transmit_data($addr,efficiency)
                  set max_temp_efficiency $max_efficiency
                  
                  if {$efficiency_flag == 0} {
                   set min_temp_efficiency [lindex $transmit_data($addr,efficiency) 0]
                   set min_efficiency $min_temp_efficiency
                   set efficiency_flag 1
                  }                 
               } 
              
               if {$transmit_data($addr,efficiency) < $min_temp_efficiency} {
                  set min_efficiency $transmit_data($addr,efficiency)
                  set min_temp_efficiency $min_efficiency
               } 
              
#               if {$transmit_data($addr,throughput) > $max_temp_throughput} {
#                  set max_throughput $transmit_data($addr,throughput)
#                  set max_temp_throughput $max_throughput
              
#                  if {$throughput_flag == 0} {
#                   set min_temp_throughput [lindex $transmit_data($addr,throughput) 0]
#                   set min_throughput $min_temp_throughput
#                   set throughput_flag 1
#                  }
#               } 
              
#               if {$transmit_data($addr,throughput) < $min_temp_throughput} {                 
#                  set min_throughput $transmit_data($addr,throughput)
#                  set min_temp_throughput $min_throughput
#               } 
              
#               puts stdout " "
               
               } else {
               
               puts stdout "Warning: receiver $addr did not finish reception!"
            }
    }
       
    puts -nonewline stdout "Avg Efficiency: [format "%3.4f" $avg_efficiency]"
    puts -nonewline $outp_fd "avg,min,max-- [format "%7.4f" $avg_efficiency]"
#    puts -nonewline stdout " Avg Throughput: [format "%7.2f" $avg_throughput]"
    puts stdout "     Total Useful Pkts: $tot_useful_recv"
    puts -nonewline stdout "Min Efficiency: [format "%3.4f" $min_efficiency]"
    puts -nonewline $outp_fd2 "useful,duplicate,linear--  [format "%12.0f" $tot_useful_recv]"
    puts -nonewline $outp_fd "  [format "%7.4f" $min_efficiency]"
#   puts -nonewline stdout " Min Throughput: [format "%7.2f" $min_throughput]"
    puts stdout "  Total Duplicate Pkts: $tot_full_disc"
    puts -nonewline stdout "Max Efficiency: [format "%3.4f" $max_efficiency]"
    puts -nonewline $outp_fd2 "[format "%12.0f" $tot_full_disc]"
    puts $outp_fd  "  [format "%7.4f" $max_efficiency]"
#  puts -nonewline stdout " Max Throughput: [format "%7.2f" $max_throughput]"
    puts stdout "     Total Linear Pkts: $tot_lin_dep_disc"
    puts $outp_fd2 "[format "%12.0f" $tot_lin_dep_disc]"
   




